/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * expr_c: expression builder for ZXSimulator, a ZX81 ROM emulator
 */


#include stdio_h
#include cmd_h

/* gets a variable from command line */
getvar (var)
   char var[];
{
   int  i;
   char ch;

   while (str[sp] == ' ') sp++;
   if (!isalpha(str[sp])) return NOVAL;
   i=0;
   while (i<VARLEN)
      if (isalpha(ch=str[sp]) || isdigit(ch) || (ch==' '))
         { if (ch != ' ')
              { var[i++] = ch; sp++; }
           else 
              { while (str[sp]==' ') sp++;
                switch (str[sp])
                   { case 'T':
                       if (getkey("THEN") == NOERR) { sp = sp - 4; goto end_var; }
                       if (getkey("TO") == NOERR) { sp = sp - 2; goto end_var; }
                       break;
                    case 'S':
                       if (getkey("STEP") == NOERR) { sp = sp - 4; goto end_var; }
                       break;
                    case 'A':
                       if (getkey("AND") == NOERR) { sp = sp - 3; goto end_var; }
                       break;
                    case 'O':
                       if (getkey("OR") == NOERR) { sp = sp - 2; goto end_var; }
         }    }    }
      else
         { if (ch == '$' || ch == '(')
              { var[i++] = ch; sp++; if (i == VARLEN) break; }
           end_var: var[i] = '\0'; return NOERR; }

   return BADARG;
}


/* finds var in hashtable */
findvar(var,vp)
   char var[]; int *vp;
{
   int i, j, *num;
   addr *nam, *ptr;

   i = hash(var);
   
   j=i;
   do {
      if (var_nam[j] == 0L)
         { *vp = j; return NOVAL; }
      else if (!strcmp(var_nam[j],var))
         { *vp = j; return NOERR; }
      j=(j+1)%var_len;
   } while (j!=i);


   /* if hash is filled up we double space */
   nam=var_nam; ptr=var_ptr; num=var_num;

   var_len+=VARSIZE;
   initvar(CREATE);

   i=0;
   while (i<(var_len-VARSIZE))
      { findvar(nam[i],&j);
        var_nam[j]=nam[i]; var_ptr[j]=ptr[i]; var_num[j]=num[i]; i++; }
   free(nam); free(ptr); free(num);

   return findvar(var,vp);
}


/* sets a variable */
storevar (var,ind,val_i,val_s)
   char var[]; int ind, val_i; char val_s[];
{
   addr mem;
   char ch;
   int  i, err;

   if ((err=findvar(var,&i)) == NOVAL)
      { memarg(&mem,strlen(var)+1);
        var_nam[i]=mem; strcpy(mem,var); }

   if ((ch=var[strlen(var)-1]) == '$')
      { memarg(&mem,strlen(val_s)+1);
        if (err == NOERR) free(var_ptr[i]);
        var_ptr[i] = mem; strcpy(mem,val_s); 
        if (ind < 0) var_num[i] = 0; }
   else if (ch == '(')
      setintp(var_ptr[i],ind,val_i);
   else
      var_num[i] = val_i;

   return NOERR;
}


/* retrievs value of a variable */
fetchvar (var,ind,val_i,val_s)
   char var[]; int ind, *val_i; addr *val_s;
{
   char ch;
   char *ptr;
   int i;

   if (findvar(var,&i) != NOERR) return BADARG;

   if ((ch=var[strlen(var)-1]) == '$')
      { doSTR=1;
        memarg(val_s,strlen(ptr=var_ptr[i])+1); strcpy(*val_s,ptr);
        *val_i = var_num[i]; }
   else 
      { if (ch == '(')
           *val_i = getintp(var_ptr[i],ind);
        else
           *val_i = var_num[i];
        memarg(val_s,NUMLEN); itoa(*val_i,*val_s); }

   return NOERR;
}


/* retrievs array index for DIM'ed array */
getindex (var,val_i,val_s)
   char var[]; int *val_i; addr *val_s;
{
   int i, j, ind, len, dim;

   if (findvar(var,&i) != NOERR) return BADARG;

   j = 1; ind = 0; len = 0; dim = getintp(var_ptr[i],0);
   while (1) {
      if (getexpr(val_i,val_s) == NOERR) free(*val_s); else return BADARG;
      len = getintp(var_ptr[i],j++);
      if (*val_i < 1 || *val_i > len) return BADARG;
      ind = ind * len + *val_i - 1;
      if (j > dim) break;
      if (getsym(",") != NOERR) return BADARG;
   }
   if (getsym(")") != NOERR) return BADARG;

   *val_i = (1 + dim) + ind;
   memarg(val_s,NUMLEN); itoa(*val_i,*val_s);

   return NOERR;
}


/* gets number from command line */
getnum (val_i,val_s)
   int *val_i; addr *val_s;
{
   char ch;
   int i;

   while (str[sp] == ' ') sp++;
   if (!isdigit(str[sp])) return NOVAL;
   i=0; memarg(val_s,NUMLEN);
   while (isdigit(ch=str[sp]))
      { setchp(*val_s,i++,ch); sp++; }
   setchp(*val_s,i,'\0');
   *val_i = atoi(*val_s);

   return NOERR;
}


/* gets quoted string from command line */
getstr (val_s)
   addr *val_s;
{
   char ch;
   int i, tp, len;

   while (str[sp] == ' ') sp++;
   if (str[sp] == '"') { doSTR = 1; tp = ++sp; } else return NOVAL;
   while((ch=str[tp])!='"')
      if (ch=='\\') tp+=3; else if (ch=='%') tp+=2; else { tp++; if (ch==EOL) return BADCHAR; }
   len = tp-sp; memarg(val_s,len+1);
   i=0; while (i<len) setchp(*val_s,i++,str[sp++]);
   setchp(*val_s,i,'\0'); sp++;

   return NOERR;   
}


/* parses substring expression */
getsubstr (tmp_i,tmp_s,val_i,val_s)
   int tmp_i; addr tmp_s;
   int *val_i; addr *val_s;
{
   char ch;
   int from, to, i, j, cp, in, idx, len;

   switch (getrange(tmp_i,&from,&to,(len=ZXlen(tmp_s))))
      { case NOVAL:
           *val_i=atoi(tmp_s); *val_s=tmp_s; return NOERR;
        case BADARG:
           free(tmp_s); return BADARG; }

   if (to < from)
      { j=0; memarg(val_s,1); }
   else
      {
        cp=0;
        while (cp<2)  /* first pass see how much space, no copy */
        {
           if (cp) memarg(val_s,j+1);

           i=0; j=0; in=0; idx=0;
           while (idx<len)
           {
              idx++;
              switch (ch=getchp(tmp_s,i))
              {
                 case '%':
                    if (in || from == idx) 
                       { if (cp) 
                            { setchp(*val_s,j++, ch); i++;
                              setchp(*val_s,j++,getchp(tmp_s,i)); i++; }
                         else
                            { i=i+2; j=j+2; }
                         if (idx == to) goto end_idx;
                         in = 1; }
                    else
                       i=i+2;
                    break;
                 case '\\':
                    if (in || from == idx)
                       { if (cp) 
                            { setchp(*val_s,j++, ch); i++;
                              setchp(*val_s,j++,getchp(tmp_s,i)); i++;
                              setchp(*val_s,j++,getchp(tmp_s,i)); i++; }
                         else
                            { i=i+3; j=j+3; }
                         if (idx == to) goto end_idx;
                         in = 1; }
                    else
                       i=i+3;
                    break;
                 default:
                    if (in || from == idx)
                       { if (cp) setchp(*val_s,j,ch);
                         j++;
                         if (idx == to) goto end_idx;
                         in = 1; }
                    i++;
              }
           }
           end_idx:
           cp++;
        }
      }

   setchp(*val_s,j,'\0'); *val_i=atoi(*val_s);
   free(tmp_s);

   return NOERR;
}


/* gets range of string */
getrange (size,from,to,len)
   int size, *from, *to, len;
{
   char *tmp_s;
   int all, row, noTO;

   if (getsym("(") != NOERR) { if (size) { return BADARG; } return NOVAL; }

   all = 0; row = 0;
   if (size)
      { if (getexpr(&row,&tmp_s) == NOERR) free(tmp_s); else return BADARG;
        if (getsym(",") != NOERR)
           { if (getsym(")") == NOERR) { all=1; sp--; } else return BADARG; }
        len = row * size; row = len - size; }

   if (noTO=(getkey("TO") != NOERR))
      { if (getexpr(from,&tmp_s) == NOERR) free(tmp_s); else goto set_from; }
   else
      { set_from: *from = 1; }

   *from += row;

   if (!noTO || (getkey("TO") == NOERR))
      { if (getexpr(to,&tmp_s) == NOERR) { *to += row; free(tmp_s); } else *to = len; }
   else
      { if (all) *to = *from + size - 1; else *to = *from; }

   if (getsym(")") != NOERR) return BADARG;

   if ((*from < 1) || (*to < 0)) return BADARG;
   if (*to < *from) { *from=0; *to=-1; } else if ((*from > len) || (*to > len)) return BADARG;

   doSTR=1;
   return NOERR;
}


/* gets value from command line */
getval (val_i,val_s)
   int *val_i; addr *val_s;
{
   addr tmp_s;
   int  i, tmp_i, err;
   char var[VARLEN];

   if (getsym("(") == NOERR)  /* eval subexpression */
      { if (getexpr(val_i,val_s) != NOERR) return BADARG;
        return getsym(")"); }
   else if ((err=evalfun(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getnum(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getstr(&tmp_s)) != NOVAL)
      { if (err != NOERR) return err;
        return getsubstr(0,tmp_s,val_i,val_s); }
   else if (getvar(var) == NOERR)
      { switch (var[strlen(var)-1])
           { case '$':
                if (fetchvar(var,0,&tmp_i,&tmp_s) == NOERR)
                   return getsubstr(tmp_i,tmp_s,val_i,val_s);
                break;
             case '(':
                if (getindex(var,&tmp_i,&tmp_s) != NOERR) { break; } free(tmp_s);
             default:
                return fetchvar(var,tmp_i,val_i,val_s); } }

   return BADARG;
}


/* gets {chr$(),len()} functions and unary {-,not} expressions */
evalfun (val_i,val_s)
   int *val_i; addr *val_s;
{
   int i, inv;

   delim = lst;  /* function delimeter different from other keywords */
   if (getsym("-") == NOERR) {
      if (getval(val_i,val_s) != NOERR) return BADARG;
      *val_i = -*val_i;
   }
   else switch (str[sp])
   {
      case 'A':
         if (getkey("ABS") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = abs(*val_i);
            break;
         }
         goto noval;
      case 'C':
         if (getkey("CHR$") == NOERR) {
            if (getval(val_i,val_s) != NOERR) { return BADARG; } free(*val_s);
            if (*val_i >> 8) return BADARG;

            i=0; inv = *val_i > 127; memarg(val_s,2+inv);
            if (inv) { *val_i = *val_i - 128; setchp(*val_s,i++,'%'); }

            if (*val_i == 0)
               *val_i = 32;
            else if (*val_i <11 || *val_i > 63)
               { *val_i = 15; i=0; }
            else if (*val_i >= 38 )
               *val_i = *val_i + 27;  /* 'A' is 65 for ASCII - 38 for ZX81 */
            else if (*val_i >= 28)
               *val_i = *val_i + 20;  /* '0' is 48 for ASCII - 28 for ZX81 */
            else
               *val_i = sym[*val_i-11];

            doSTR = 1; setchp(*val_s,i++,*val_i); setchp(*val_s,i,'\0');
            goto skip_itoa;  /* we want to keep string value */
         }
         if (getkey("CODE") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 0; *val_i = getchp(*val_s,0);
      
            if (*val_i == '\0')
               *val_i = 0;
            else if (*val_i == '\\')
               { if (getchp(*val_s,1)==':' && getchp(*val_s,2)==':') *val_i=128; else *val_i=67; }
            else
               { if (*val_i != '%') inv = 0; else { inv = 128; *val_i = getchp(*val_s,1); }

                 if (*val_i == 32)
                    *val_i = 0 + inv;
                 else if (*val_i >= 48 && *val_i <= 57)
                    *val_i = *val_i - 20 + inv;   /* '0' is 48 for ASCII - 28 for ZX81 */
                 else if (*val_i >= 65 && *val_i <= 90)
                    *val_i = *val_i - 27 + inv;   /* 'A' is 65 for ASCII - 38 for ZX81 */
                 else if (*val_i < 192)  /* 192 & up are QL control characters we want */
                    { i=0; while (i<17) if (sym[i] == *val_i) break; else i++;
                      if (i == 17) *val_i = 67; else *val_i = i + 11 + inv; } }
            break;
         }
         goto noval;
      case 'I':
         if (getkey("INKEY$") == NOERR) {
            i = 0; memarg(val_s,2); trap3(key, ret);
            if (!ret[0]) { setchp(*val_s,i++,toupper(getchar())); } setchp(*val_s,i,'\0');
            if (getchp(*val_s,0)==32) { setchp(*val_s,0,252); }  /* for space 32 map to 252 */
            doSTR = 1; *val_i = atoi(*val_s);
            goto skip_itoa;  /* want to keep string value */
         }
         if (getkey("INT") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            break;
         }
         goto noval;
      case 'L':
         if (getkey("LEN") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 0; *val_i = ZXlen(*val_s);
            break;
         }
         goto noval;
      case 'N':
         if (getkey("NOT") == NOERR) {
            if (getexpr(val_i,val_s) != NOERR) return BADARG;
            *val_i = !*val_i;
            break;
         }
         goto noval;
      case 'P':
         if (getkey("PI") == NOERR) {
            *val_i = PI;
            goto skip_free;
         }
         goto noval;
      case 'R':
         if (getkey("RND") == NOERR) {
            seed = ((seed * P1) + P2) % N;
            *val_i = seed; *val_i = abs(*val_i);   /* int from long then its absolute value */
            if (getsym("*") != NOERR) goto skip_free;
            if (getval(&i,val_s) != NOERR) return BADARG;
            *val_i = *val_i % i; if (i<0) *val_i = -*val_i;    /* hack until float */
            break;
         }
         goto noval;
      case 'S':
         if (getkey("SGN") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = sign(*val_i);
            break;
         }
         if (getkey("STR$") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 1;
            goto skip_itoa;  /* we want to keep string value */
         }
         goto noval;
      case 'V':
         if (getkey("VAL") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 0;
            break;
         }
      default:
         noval:
         delim = spc; return NOVAL;
   }

   /* free string since function only returns numeric value */
   free(*val_s); skip_free:
   memarg(val_s,NUMLEN); itoa(*val_i,*val_s); skip_itoa:

   delim = spc;  /* reset delimeter */
   return NOERR;
}


/* parses ** expressions */
evalpow (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i, pow, i;
   addr right_s;

   if (getval(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            if (str[sp+1] != '*') return NOERR;
            sp = sp + 2;
            if (getval(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            i = 0; pow = 1;
            while (i++ < right_i)
               pow *= *left_i;

            *left_i = pow;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free (*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}


/* parses {*,/} expressions */
evalmul (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evalpow(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i * right_i;
            break;
         }
         case '/' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i / right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}


/* parses {+,-} expressions */
evaladd (left_i,left_s)
   int *left_i; addr *left_s;
{
   int i, j, right_i;
   addr right_s, tmp_s;

   if (evalmul(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      j=0; 
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '+' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }

            /* concatenate string if we grabbed one */
            if (doSTR)
               { memarg(&tmp_s,strlen(*left_s)+strlen(right_s)+1);
                 i=0; while (getchp(*left_s,i)!='\0') setchp(tmp_s,j++,getchp(*left_s,i++));
                 i=0; while (getchp(right_s,i)!='\0') setchp(tmp_s,j++,getchp(right_s,i++));
                 setchp(tmp_s,j,'\0'); free(*left_s); *left_s = tmp_s;
                 *left_i = atoi(*left_s); }
            else
               *left_i = *left_i + right_i;
            break;
         }
         case '-' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i - right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(right_s);
      if (!doSTR)
         { free(*left_s); memarg(left_s,NUMLEN); itoa(*left_i,*left_s); }
   }
}


/* parses {=,>,<,<>,<=,>=} expressions */
evaleql (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   doSTR = 0;  /* used for string casting */
   if (evaladd(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getsym("=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) == 0);
         else
            *left_i = *left_i == right_i;
      }
      else if (getsym("<=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) <= 0);
         else
            *left_i = *left_i <= right_i;
      }
      else if (getsym("<>") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) != 0);
         else
            *left_i = *left_i != right_i;
      }
      else if (getsym("<") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) < 0);
         else
            *left_i = *left_i < right_i;
      }
      else if (getsym(">=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) >= 0);
         else
            *left_i = *left_i >= right_i;
      }
      else if (getsym(">") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) > 0);
         else
            *left_i = *left_i > right_i;
      }
      else {
         return NOERR;
      }
      doSTR = 0;
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}


/* parses {and,or} expressions */
evalbool (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i, leftSTR;
   addr right_s;

   if (evaleql(left_i,left_s) != NOERR) return BADARG;
   leftSTR = doSTR;
   while (1)
   {
      if (getkey("AND") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) goto err_left;
         if (doSTR) goto err_both;            /* ZX81 BASIC: can't have string in right arg */
         if (!right_i) *left_i = 0; else goto skip_left;   /* returns left arg if both true */
         if (leftSTR) { setchp(*left_s,0,'\0'); goto skip_left; }  /* false is empty string */
      }
      else if (getkey("OR") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) goto err_left;
         if (leftSTR || doSTR) goto err_both;   /* ZX81 BASIC: can't have string in OR args */
         if (right_i) *left_i = 1;                /* returns 1 if right true, else left arg */ 
      }
      else {
         return NOERR;

         /* stick error cleanup here to reduce repeat code */
         err_both: free(right_s);
         err_left: free(*left_s);
         return BADARG;
      }
      free(*left_s); memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
      skip_left: free(right_s);
   }
}
